%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
%}

%union {
  double d;
  char* s;
  struct expr* expr;
  struct print* print;
  struct varlist* varlist;
  struct inst* inst;
}

%{

  typedef struct stack {
    double val;
    struct stack* next;
  } stack;

  typedef struct expr {
    int    type;
    
    char*  var;         /* for VARIABLE type */
    double d;           /* for NOMBRE type */

    struct expr* left;  /* for unary and binary nodes */
    struct expr* right; /* for binary nodes */
    struct expr* next;  /* for expression lists */
  } expr;
  
  typedef struct varlist {
    char* var;
    stack* stack;
    struct varlist* next;
  } varlist;
  
  typedef struct inst {
    int type;
    
    expr* expr1;          /* test in IF, WHILE */
    expr* expr2;          /* arguments de COLOR */
    expr* expr3;          
    struct inst*  block1; /* block in IF, WHILE and FOR */
    struct inst*  block2; /* then block in IF */
    char* s;              /* string for PRINT */
    char* var;            /* variable in ASSIGN */
    varlist* varlist;     /* for procedure definition */
    
    struct inst* next;    /* next instruction if in a bloc */
    
  } inst;


  expr* expr_bin(int type, expr* e1, expr* e2){
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=type; e->left=e1; e->right=e2;
    return e;
  }

  expr* expr_nb(double d) {
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=NOMBRE; e->d=d;
    return e;
  }

  expr* expr_var(char* var) {
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=VARIABLE; e->var=var;
    return e;
  }

  expr* expr_input() {
    expr* e = (expr*)malloc(sizeof(expr));
    e->type=INPUT;
    return e;
  }

  inst* inst_while(expr* e1, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=WHILE; i->expr1=e1; i->block1=block;
    return i;
  }

  inst* inst_if(expr* e1, inst* block1, inst* block2) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=IF; i->expr1=e1; i->block1=block1; i->block2=block2;
    return i;
  }

  inst* inst_print(expr* e1,char* s) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PRINT; i->expr1=e1; i->s=s;
    return i;
  }

  inst* inst_assign(char* var, expr* e1) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=ASSIGN; i->var=var; i->expr1=e1;
    return i;
  }

  inst* inst_simple(int type, expr* e1, expr* e2, expr* e3) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=type; i->expr1=e1; i->expr2=e2; i->expr3=e3;
    return i;
  }

  inst* inst_proc(char* var, varlist* arg, inst* block) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PROC; i->var=var; i->varlist=arg; i->block1=block;
    return i;
  }

  inst* inst_call(char* var, expr* arg) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=VARIABLE; i->var=var; i->expr1=arg;
    return i;
  }

  varlist* varlist_var(char* var) {
    varlist* v = (varlist*)malloc(sizeof(varlist));
    v->stack = NULL;
    v->var = var;
    return v;
  }

  inst* print_chaine(char* s) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PRINT; i->s=s; i->expr1=NULL;
    return i;
  }

  inst* print_expr(expr* expr) {
    inst* i = (inst*)malloc(sizeof(inst));
    i->type=PRINT; i->expr1=expr; i->s=NULL;
    return i;
  }


  typedef struct var {
    char* name;
    double val;
    struct var* next;
  } var;
  var vars = { "", 0, NULL };
 
  var* get_var(char *s) {
    var* v;
    for (v=&vars;v;v=v->next) 
      if (!strcmp(v->name,s)) return v;
    for (v=&vars;v->next;v=v->next);
    v->next=(var*)malloc(sizeof(var));
    v->next->name = s;
    v->next->next = NULL;
    return v->next;
  }

  double get_val(char* s) {
    var* v;
    for (v=&vars;v;v=v->next) if (!strcmp(v->name,s)) return v->val;
    return 0;
  }

  typedef struct proc {
    char* name;
    varlist* args;
    inst* block;
    struct proc* next;
  } proc;
  proc procs = {"",NULL,NULL,NULL};

  FILE* graph;
  double x=594/2,y=838/2,ang=0;

  double eval_expr(expr* e) {
    switch (e->type) {
    case INPUT:
      { 
	double d;
	scanf("%le",&d);
        return d;
      }
    case NOMBRE: return e->d;
    case PLUS: return eval_expr(e->left)+eval_expr(e->right);
    case MOINS: return eval_expr(e->left)-eval_expr(e->right);
    case MULTIPLIE: return eval_expr(e->left)*eval_expr(e->right);
    case DIVISE: return eval_expr(e->left)/eval_expr(e->right);
    case OPPOSE: return -eval_expr(e->left);
    case EGAL: return (eval_expr(e->left)==eval_expr(e->right) ? 1 : 0);
    case DIFFERENT: return (eval_expr(e->left)!=eval_expr(e->right) ? 1 : 0);
    case SUPERIEUR: return (eval_expr(e->left)>eval_expr(e->right) ? 1 : 0);
    case SUPERIEUR_EGAL: return (eval_expr(e->left)>=eval_expr(e->right) ? 1 : 0);
    case INFERIEUR: return (eval_expr(e->left)<eval_expr(e->right) ? 1 : 0);
    case INFERIEUR_EGAL: return (eval_expr(e->left)<=eval_expr(e->right) ? 1 : 0);
    case VARIABLE: return get_val(e->var);
    case ET: return (eval_expr(e->left) && eval_expr(e->right) ? 1 : 0);
    case OU: return (eval_expr(e->left) || eval_expr(e->right) ? 1 : 0);
    case NON: return (eval_expr(e->left) ? 1 : 0);

    default: 
      printf("\nInternal error: case %i in exec_expr\n\n",e->type); 
      exit(1);
    }
  }

  void exec_block(inst* i) {
    while (i) {
      switch (i->type) {
	
      case ASSIGN:
	get_var(i->var)->val = eval_expr(i->expr1);
	break;	

      case PRINT: {
	if (i->s) printf("%s",i->s);
	if (i->expr1) printf("%f",eval_expr(i->expr1));
	break;
      }

      case WHILE: 
	while (eval_expr(i->expr1)) exec_block(i->block1);
	break;

      case IF:
	if (eval_expr(i->expr1)) exec_block(i->block1);
	else if (i->block2) exec_block(i->block2);
	break;

      case PROC: {
	proc* p;
	for (p=&procs;p->next;p=p->next);
	p->next=(proc*)malloc(sizeof(proc));
	p->next->name  = i->var;
	p->next->block = i->block1;
	p->next->args  = i->varlist;
	p->next->next  = NULL;
	break;
      }

      case VARIABLE: {
 	proc* p;
	varlist* l;
	expr* e;
	for (p=&procs;p;p=p->next)
	  if (!strcmp(p->name,i->var)) break;
	if (!p) { printf("Procedure %s non definie!\n",i->var); exit(1); }

	for (e=i->expr1,l=p->args;e || l;e=e->next,l=l->next)
	  if (!l || !e) { printf("Mauvais nombre d'arguments dans l'appel de procedure %s!\n",i->var); exit(1); }
	  else {
	    stack *s = (stack*)malloc(sizeof(stack));
	    s->next = l->stack;
	    s->val = eval_expr(e);
	    l->stack = s;
	  }

	for (l=p->args;l;l=l->next) {
	  double d;
	  stack* s = l->stack;
	  var* v = get_var(l->var);
	  d = v->val; v->val=s->val; s->val = d;
	}

	exec_block(p->block);
	
	for (l=p->args;l;l=l->next) {
	  stack* s = l->stack;
	  var* v = get_var(l->var);
	  v->val = s->val;
	  l->stack = s->next;
	  free(s);
	}
	break;
      }

      case TD: ang-=eval_expr(i->expr1); break;
      case TG: ang+=eval_expr(i->expr1); break;	

      case AV: {
	double a = eval_expr(i->expr1);
	double xx = x+a*cos(ang/180.*3.1415927);
	double yy = y+a*sin(ang/180.*3.1415927);
	fprintf(graph,"%f %f moveto %f %f lineto stroke\n",x,y,xx,yy);
	x=xx; y=yy;
	break;
      }

      case COLOR:
	fprintf(graph,"%f %f %f setrgbcolor\n",
		eval_expr(i->expr1),eval_expr(i->expr2),eval_expr(i->expr3));
	break;

      default: 
	printf("\nInternal error: case %i in exec_block\n\n",i->type); 
	exit(1);
      }
      i=i->next;
    } 
  }
  
  void prog(inst* i) {
    graph = fopen("dessin.ps","w");
    fprintf(graph,"%%!postscript\n");
    exec_block(i);
    printf("\n");
    fprintf(graph,"showpage\n");
    fclose(graph);
  }

%}

%token PRINT INPUT IF ELSE WHILE ASSIGN PROC
%token PLUS MOINS MULTIPLIE DIVISE
%token ET OU NON
%token SUPERIEUR SUPERIEUR_EGAL INFERIEUR INFERIEUR_EGAL DIFFERENT EGAL
%token PARENTHESE_GAUCHE PARENTHESE_DROITE VIRGULE POINTVIRGULE
%token BLOC_GAUCHE BLOC_DROITE
%token ERROR QUIT
%token NOMBRE CHAINE VARIABLE
%token AV TG TD COLOR

%left ET
%left OU
%left NON
%left SUPERIEUR SUPERIEUR_EGAL INFERIEUR INFERIEUR_EGAL DIFFERENT EGAL
%left PLUS MOINS
%left MULTIPLIE DIVISE
%nonassoc OPPOSE 

%start prog

%type<inst> instrliste instr block
%type<expr> expr exprlist
%type<d> NOMBRE
%type<s> VARIABLE CHAINE
%type<varlist> varlist

%%

prog:   instrliste QUIT { prog($1); YYACCEPT; }
;

instrliste:  { $$=NULL; }
| instr instrliste { $$=$1; $$->next=$2; }
;

instr: IF PARENTHESE_GAUCHE expr PARENTHESE_DROITE block { $$=inst_if($3,$5,NULL); }
| IF PARENTHESE_GAUCHE expr PARENTHESE_DROITE block ELSE block { $$=inst_if($3,$5,$7); }
| VARIABLE ASSIGN expr POINTVIRGULE { $$=inst_assign($1,$3); }
| WHILE PARENTHESE_GAUCHE expr PARENTHESE_DROITE block { $$=inst_while($3,$5); }
| PROC VARIABLE PARENTHESE_GAUCHE varlist PARENTHESE_DROITE block { $$=inst_proc($2,$4,$6); }
| VARIABLE PARENTHESE_GAUCHE exprlist PARENTHESE_DROITE POINTVIRGULE { $$=inst_call($1,$3); }
| AV PARENTHESE_GAUCHE expr PARENTHESE_DROITE POINTVIRGULE { $$=inst_simple(AV,$3,NULL,NULL); }
| TD PARENTHESE_GAUCHE expr PARENTHESE_DROITE POINTVIRGULE { $$=inst_simple(TD,$3,NULL,NULL); }
| TG PARENTHESE_GAUCHE expr PARENTHESE_DROITE POINTVIRGULE { $$=inst_simple(TG,$3,NULL,NULL); }
| COLOR PARENTHESE_GAUCHE expr VIRGULE expr VIRGULE expr PARENTHESE_DROITE POINTVIRGULE { $$=inst_simple(COLOR,$3,$5,$7); }
| PRINT PARENTHESE_GAUCHE expr PARENTHESE_DROITE POINTVIRGULE { $$=print_expr($3); }
| PRINT PARENTHESE_GAUCHE CHAINE PARENTHESE_DROITE POINTVIRGULE { $$=print_chaine($3); }
;

block:  BLOC_GAUCHE instrliste BLOC_DROITE { $$=$2; }
;

expr:  VARIABLE { $$=expr_var($1); }
| expr PLUS expr { $$=expr_bin(PLUS,$1,$3); }
| expr MOINS expr  { $$=expr_bin(MOINS,$1,$3); }
| expr MULTIPLIE expr  { $$=expr_bin(MULTIPLIE,$1,$3); }
| expr DIVISE expr  { $$=expr_bin(DIVISE,$1,$3); }
| expr SUPERIEUR expr  { $$=expr_bin(SUPERIEUR,$1,$3); }
| expr SUPERIEUR_EGAL expr  { $$=expr_bin(SUPERIEUR_EGAL,$1,$3); }
| expr INFERIEUR expr  { $$=expr_bin(INFERIEUR,$1,$3); }
| expr INFERIEUR_EGAL expr  { $$=expr_bin(INFERIEUR_EGAL,$1,$3); }
| expr DIFFERENT expr  { $$=expr_bin(DIFFERENT,$1,$3); }
| expr EGAL expr  { $$=expr_bin(EGAL,$1,$3); }
| expr ET expr { $$=expr_bin(ET,$1,$3); }
| expr OU expr { $$=expr_bin(OU,$1,$3); }
| NON expr     { $$=expr_bin(NON,$2,NULL); }
| MOINS %prec OPPOSE expr  { $$=expr_bin(OPPOSE,$2,NULL); }
| PARENTHESE_GAUCHE expr PARENTHESE_DROITE { $$=$2; }
| NOMBRE { $$=expr_nb($1); }
| INPUT { $$=expr_input(); }
;

exprlist: expr { $$=$1; $$->next=NULL; }
| expr VIRGULE exprlist { $$=$1; $$->next=$3; }
;

varlist: VARIABLE { $$=varlist_var($1); $$->next=NULL; }
| VARIABLE VIRGULE varlist { $$=varlist_var($1); $$->next=$3; }
;

%%

int main (int argc, char** argv)
{
  if (argc==2) {
    FILE* f = fopen(argv[1],"r");
    if (f!=NULL) {
      yy_switch_to_buffer(yy_create_buffer(f,4096));
      yyparse();
      fclose(f);
      return 0;
    }
    printf("Fichier %s introuvavle!\n",argv[1]);
  }
  printf("interpreteur programme\n");
  return 1;
}
